#ifndef LZMA_VERSION_H
#define LZMA_VERSION_H

/*
    Version and copyright information used by LZMA utils.
*/

static const char *LZMA_SDK_VERSION_STRING = "4.32";

static const char *LZMA_SDK_COPYRIGHT_STRING =
		"Copyright (C) 1999-2005 Igor Pavlov";

static const char *LZMA_SDK_COPYRIGHT_INFO =
		"  See http://7-zip.org/sdk.html or the documentation of LZMA SDK for\n"
		"  the license. For reference, the version 4.32 is free software\n"
		"  licensed under the GNU LGPL.";


static const char *LZMA_UTILS_VERSION_STRING = PACKAGE_VERSION;

static const char *LZMA_UTILS_COPYRIGHT_STRING =
		"Copyright (C) 2005 Lasse Collin";

static const char *LZMA_UTILS_COPYRIGHT_INFO =
		"This program comes with ABSOLUTELY NO WARRANTY.\n"
		"You may redistribute copies of this program\n"
		"under the terms of the GNU General Public License.\n"
		"For more information about these matters, see the file "
		"named COPYING.\n";

#endif /* ifndef LZMA_VERSION_H */
