// Common/Types.h

#ifndef __COMMON_TYPES_H
#define __COMMON_TYPES_H

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <inttypes.h>

typedef uint8_t Byte;
typedef int_fast16_t Int16;
typedef uint_fast16_t UInt16;
typedef int32_t Int32;
typedef uint32_t UInt32;
typedef int64_t Int64;
typedef uint64_t UInt64;

#endif
